/*
 * EduQiz - a computer-assisted learning tool to help students or children
 *          in tasks that require memorization of words, dates, etc.
 *
 * EduQiz is published under the terms of the MIT license as stated below.
 *
 * Copyright (C) 2022-2023 Mateusz Viste
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/* TESTBUILD is a way of using a modern C compiler as a static analisis tool,
 * the build won't yield a working executable */
#ifdef TESTBUILD
  #define MK_FP(x, y) (void *)(x * 16 + y)
  #define far
  #define FNM_IGNORECASE 0
  #define FP_SEG(x) (unsigned long)(x)
  #define FP_OFF(x) (unsigned long)(x)
  #define _dos_getvect(x) (void *)(x)
#else
  #include <dos.h>    /* _dos_getvect() */
  #include <i86.h>    /* MK_FP() */
#endif


#include <fnmatch.h>
#include <malloc.h>
#include <stdio.h>  /* sprintf() */
#include <stdlib.h> /* getenv() */
#include <string.h>
#include <strings.h> /* strcasecmp(), bzero() */

#include "mdr/bios.h"
#include "mdr/cout.h"
#include "mdr/keyb.h"
#include "mdr/pcx.h"
#include "mdr/sbdigi.h"
#include "mdr/trigint.h"
#include "mdr/unzip.h"
#include "mdr/video.h"
#include "mdr/wave.h"

#define PVER "2023"
#define PDATE "2022-2023"

/* Color indexes of the UI palette, as defined in the MONKEY.PCX file */
#define PALID_BLACK        240
#define PALID_RED_DARK    (240+1)
#define PALID_BLACK_LIGHT (240+2)
#define PALID_BROWN       (240+6)
#define PALID_GRAY_DARK   (240+9)
#define PALID_BEIGE_DARK  (240+11)
#define PALID_GREEN       (240+12)
#define PALID_BEIGE_LIGHT (240+13)
#define PALID_GRAY_LIGHT  (240+14)
#define PALID_WHITE       (240+15)


struct conf {
  const char *qizfname;
  unsigned long titlepcx_offset;
  unsigned short titlepcx_len;
  unsigned char nosound;
  unsigned char FFU; /* for future use */
  FILE *fd;
  unsigned short qizcount; /* total amount of questions in the quiz */
  unsigned short qizlimit; /* limit of questions during a single quizz session */
  unsigned short stats_answers_total; /* count of validated quiz questions */
  unsigned short stats_answers_good;  /* count of validated quiz questions answered correctly at first try */
  struct qizlist {
    unsigned long off;
    unsigned short len;
    unsigned char flags;
  } list[1];
};

#define QIZFLAG_VALIDATED 1
#define QIZFLAG_HINTED 2


struct qiz {
  unsigned long pcxoffset;
  unsigned long wavoffset;
  unsigned long txtoffset;
  unsigned short pcxlen;
  unsigned short wavlen;
  unsigned short txtlen;
};


struct mbuf {
  struct pcx_hdr monkey_hdr;
  char solution[16000];
  unsigned char monkeybuf[8000];
  unsigned char pal[256*3];
};


enum MONKEY_FACE {
  MONKEY_BASIC = 0,
  MONKEY_GRIN = 1,
  MONKEY_TONGUE = 2,
  MONKEY_SAD = 3,
  MONKEY_UNSURE = 4,
  MONKEY_SURPRISED = 5,
  MONKEY_OOUCH = 6,
  MONKEY_YEAH = 7
};


static int pcx_display_fullscreen(FILE *fd, unsigned long offset, unsigned short len, unsigned char palette_lastid, unsigned char fadein, unsigned char *pal) {
  struct pcx_hdr h;
  unsigned short i;
  unsigned short palid;

  /* lookup PCX properties and load palette */
  if (pcx_anal(&h, fd, offset, len) != 0) return(-1);

  /* set palette to all-black */
  video_waitvblank();
  for (palid = 0; palid <= palette_lastid; palid++) video_setpalette(palid, 0, 0, 0);

  /* load the PCX data to screen */
  pcx_load(MK_FP(0xA000, 0), 64000u, &h, fd, offset);
  pcx_convto8bpp(MK_FP(0xA000, 0), &h);

  /* apply palette (possibly with fade-in) */
  i = 0;
  if (fadein == 0) i = 32;
  for (; i <= 32; i++) {
    video_waitvblank();
    video_waitvblank();
    for (palid = 0; palid <= palette_lastid; palid++) {
      video_setpalette(palid, (h.pal[palid].r * i) / 128, (h.pal[palid].g * i) / 128, (h.pal[palid].b * i) / 128);
    }
  }

  /* copy palette if asked for */
  if (pal != NULL) memcpy(pal, h.pal, (palette_lastid + 1) * 3);

  return(0);
}


/* apply a fade-out effect, pal must contain the currently loaded VGA palette */
static void fadeout(const unsigned char *pal) {
  unsigned char i;
  unsigned char palid;
  const unsigned char *rgb;

  /* fade out effect */
  for (i = 31; i < 32; i--) {
    palid = 0;
    video_waitvblank();
    rgb = pal;
    do {
      video_setpalette(palid, (rgb[0] * i) / 128, (rgb[1] * i) / 128, (rgb[2] * i) / 128);
      palid++;
      rgb += 3;
    } while (palid != 0);
  }
}


static void display_monkey_custom(enum MONKEY_FACE id, const struct pcx_hdr *h, const unsigned char *buff, unsigned char scale, unsigned short xpos, unsigned short ypos) {
#define MONKEY_FACE_H 32
#define MONKEY_MOUTH_H 10
  unsigned short x, y;
  unsigned short pcx_scanline_offset = 0;
  unsigned short pcx_height_limit = MONKEY_FACE_H;
  unsigned char far *VGA = MK_FP(0xA000, 0);
  unsigned char pixbuf[2];
  unsigned char sx, sy;

  if (id > 0) {
    pcx_scanline_offset = MONKEY_FACE_H + (id - 1) * MONKEY_MOUTH_H;
    pcx_height_limit = MONKEY_MOUTH_H;
    ypos += 20 * scale;
  }

  /* blit monkey sprite (4bpp) to the 320x200 screen */
  for (y = 0; y < pcx_height_limit; y++) {
    for (x = 0; x <= h->max_x; x += 2) {
      pixbuf[0] = (buff[(pcx_scanline_offset+y) * h->bytes_per_scanline + (x >> 1)] >> 4) + 240;
      pixbuf[1] = (buff[(pcx_scanline_offset+y) * h->bytes_per_scanline + (x >> 1)] & 15) + 240;
      for (sy = 0; sy < scale; sy++) {
        for (sx = 0; sx < scale; sx++) {
          if (pixbuf[0] != PALID_GREEN) {
            VGA[(ypos + y*scale + sy) * 320 + xpos + (x * scale) + sx] = pixbuf[0];
          }
          if ((x < h->max_x) && (pixbuf[1] != PALID_GREEN)) {
            VGA[(ypos + y*scale + sy) * 320 + xpos + (x+1) * scale + sx] = pixbuf[1];
          }
        }
      }
    }
  }
}


static void display_monkey(enum MONKEY_FACE id, const struct pcx_hdr *h, const unsigned char *buff) {
  display_monkey_custom(id, h, buff, 1, 287, 134);
}


/* find a matching filename in a zip file and fills z with item's props.
 * this function preserves the file pointer position on the fd file descriptor
 * returns 0 on succes, non-zero otherwise */
static int ziplist_find(struct mdr_zip_item *z, FILE *fd, const char *pattern) {
  fpos_t filepos;
  int r = 0;

  /* zero out mdr_zip_item and remember where the file pointer is */
  bzero(z, sizeof(*z));
  if (fgetpos(fd, &filepos) != 0) return(-1);

  while (mdr_zip_iter(z, fd) == 0) {
    if (fnmatch(pattern, z->fname, FNM_IGNORECASE) == 0) goto CLEANUP;
  }

  /* no match found */
  r = -1;

  CLEANUP:
  fsetpos(fd, &filepos);

  return(r);
}


/* opens the dat file and returns a pointer to it, setting offset and siz accordingly */
static FILE *locateasset(const char *pattern, unsigned long *offset, unsigned short *siz) {
  FILE *fd = NULL;
  struct mdr_zip_item z;

  fd = fopen("EDUQIZ.DAT", "rb");
  if (fd == NULL) goto OOPS;

  if (ziplist_find(&z, fd, pattern) != 0) goto OOPS;

  /* must be a file (not a dir), not compressed, not encrypted and must be less
   * than 65536 bytes */
  if (z.flags & (ZIP_FLAG_ISADIR | ZIP_FLAG_ENCRYPTED)) goto OOPS;
  if (z.compressedfilelen > 0xffff) goto OOPS;
  if (z.compressedfilelen != z.filelen) goto OOPS;

  /* success */
  *offset = z.dataoffset;
  *siz = z.compressedfilelen;
  return(fd);

  OOPS:
  if (fd != NULL) fclose(fd);
  return(NULL);
}


static int monkey_load(struct pcx_hdr *h, unsigned char *buff, unsigned short buffsz) {
  FILE *fd;
  int res = -1;
  unsigned long foff;
  unsigned short fsiz;

  fd = locateasset("MONKEY.PCX", &foff, &fsiz);
  if (fd == NULL) goto OOPS;

  if (pcx_anal(h, fd, foff, fsiz) != 0) goto OOPS;
  if (pcx_load(buff, buffsz, h, fd, foff) != 0) goto OOPS;

  res = 0; /* success */

  OOPS:
  if (fd != NULL) fclose(fd);
  return(res);
}


static void vga_outchar(unsigned char x, unsigned char y, unsigned char c, unsigned char colfg, unsigned char colbg, unsigned char transp) {
  unsigned short xx, yy;
  unsigned char far *vga = MK_FP(0xA000, 0);
  unsigned char far *fn;

  fn = (void far *)_dos_getvect(0x43); /* font data lies under the 0x43 vector, this requires the 8x16 font to have been previously selected, otherwise the default 8x8 will be there */
  fn += c * 16;

  for (xx = 0; xx < 8; xx++) {
    for (yy = 0; yy < 16; yy++) {
      unsigned char pixel = (fn[yy] >> (7 - xx)) & 1;
      unsigned char colr = colbg;
      if (pixel != 0) {
        colr = colfg;
      } else if (transp) {
        continue;
      }
      vga[((y * 16 + yy) * 320) + (x * 8) + xx] = colr;
    }
  }
}


static void vga_outstr(unsigned char x, unsigned char y, const char *s, unsigned char colfg, unsigned char colbg, unsigned char transp) {
  while (*s != 0) {
    vga_outchar(x, y, *s, colfg, colbg, transp);
    s++;
    x++;
  }
}


static void randinit(void) {
  unsigned short seed = 0;
#ifndef TESTBUILD
  _asm {
    mov ah, 0 /* get system time, number of ticks since midnight in CX:DX */
    int 0x1A
    xor dx, cx
    mov seed, dx
  }
#endif
  srand(seed);
}


/* generate a triangle wave in buffer and returns the number of samples */
static unsigned short gen_trianglewave(void *buf, short volume, short step, unsigned short maxlen) {
  unsigned short i, reallen = 0;
  short s = 128;

  for (i = 0; i < maxlen;) {
    for (; (s < (128 + volume)) && (i < maxlen); s += step) {
      ((unsigned char *)buf)[i++] = s;
      if (s == 128) reallen = i;
    }
    for (; (s > (128 - volume)) && (i < maxlen); s -= step) {
      ((unsigned char *)buf)[i++] = s;
      if (s == 128) reallen = i;
    }
  }

  return(reallen);
}


static void beep_goodanswer(struct sbdigi_ctx *sbctx, void *buf) {
  unsigned short reallen;
  /* "beep" */
  reallen = gen_trianglewave(buf, 11, 3, 900);
  /* silence */
  memset((char *)buf + reallen, 128, 100);
  reallen += 100;
  /* "bop" */
  reallen += gen_trianglewave((char *)buf + reallen, 11, 9, 900);
  /* play */
  sbdigi_playsample(sbctx, buf, reallen);
}


static void beep_badanswer(struct sbdigi_ctx *sbctx, void *buf) {
  unsigned short reallen;
  reallen = gen_trianglewave(buf, 92, 6, 2400);
  sbdigi_playsample(sbctx, buf, reallen);
}


static void play_audiofile(struct sbdigi_ctx *sbctx, void *soundbuf, unsigned short soundbufsz, const struct conf *cfg, const struct qiz *q) {
  struct wave_t wav;

  if (q->wavlen == 0) return;
  fseek(cfg->fd, q->wavoffset, SEEK_SET);
  if (wave_anal(&wav, cfg->fd) != 0) return;

  /* validate format etc (PCM, 8 bit, mono, 8000 Hz) */
  if (wav.format != 1) return;   /* 1=PCM */
  if (wav.bitdepth != 8) return;
  if (wav.channels != 1) return;
  if (wav.rate != 8000) return;

  /* ignore PCM data that does not fit in buffer */
  if (wav.len > soundbufsz) return;

  /* read PCM data */
  fseek(cfg->fd, q->wavoffset + wav.dataoffset, SEEK_SET);
  fread(soundbuf, 1, wav.len, cfg->fd);

  /* playback time */
  sbdigi_playsample(sbctx, soundbuf, wav.len);
}


static void printhelp(void) {
  mdr_coutraw_puts("EduQiz rel " PVER " ; Copyright (C) " PDATE " Mateusz Viste");
  mdr_coutraw_puts("");
  mdr_coutraw_puts("Usage: EDUQIZ [/nosound] [/max=NNN] [FILE]");
  mdr_coutraw_puts("");
  mdr_coutraw_puts("/nosound   disables sound");
  mdr_coutraw_puts("/max=NNN   limits the amount of questions to NNN (default=30)");
  mdr_coutraw_puts("");
  mdr_coutraw_puts("When no *.qiz file is provided, then DEFAULT.QIZ is loaded.");
}


/* reads a CFG entry and translate it into a config memory node */
static int load_qiz_item(struct qiz *q, struct conf *cfg, unsigned short cwiczenie, void *bufptr, unsigned short buffsz, char *sol, size_t solsz) {
  unsigned short linelen;
  unsigned short fsz;
  struct mdr_zip_item z;
  unsigned char readingsol = 0; /* "reading solution" flag */
  char *solorg = sol;
  char *buff = bufptr;

  bzero(q, sizeof(*q));
  bzero(sol, solsz);

  /* adjust solsz so it is guaranteed to end with a nul terminator */
  solsz--;

  fseek(cfg->fd, cfg->list[cwiczenie].off, SEEK_SET);
  fsz = cfg->list[cwiczenie].len;

  /* read assets files */
  for (;;) {
    /* read next line */
    for (linelen = 0; fsz > 0;) {
      if (linelen >= buffsz - 1) linelen--;
      buff[linelen++] = fgetc(cfg->fd);
      fsz--;
      if (buff[linelen - 1] == '\r') linelen--;
      if (buff[linelen - 1] == '\n') {
        linelen--;
        break;
      }
    }
    buff[linelen] = 0;

    /* empty line? move to solution loading */
    if (linelen == 0) {
      readingsol = 1;
      /* end of stream? */
      if (fsz == 0) {
        /* quit with error if solution empty */
        if (solorg[0] == 0) return(-1);
        return(0);
      }
      continue;
    }

    /* reading solution? */
    if (readingsol != 0) {
      unsigned short bufflen = strlen(buff);
      /* error out if solution too long */
      if (bufflen >= solsz) return(-1);
      /* otherwise copy it to string */
      memcpy(sol, buff, bufflen + 1);
      /* advance sol for a possible next answer */
      sol += bufflen + 1;
      solsz -= bufflen + 1;
      continue;
    }

    /* otherwise is this an existing file? */
    if (ziplist_find(&z, cfg->fd, buff) != 0) continue; /* no such file, ignore */

    /* detect the type of file */
    if (fnmatch("*.PCX", z.fname, FNM_IGNORECASE) == 0) {
      q->pcxoffset = z.dataoffset;
      q->pcxlen = z.compressedfilelen;
    }
    if (fnmatch("*.TXT", z.fname, FNM_IGNORECASE) == 0) {
      q->txtoffset = z.dataoffset;
      q->txtlen = z.compressedfilelen;
    }
    if (fnmatch("*.WAV", z.fname, FNM_IGNORECASE) == 0) {
      q->wavoffset = z.dataoffset;
      q->wavlen = z.compressedfilelen;
    }
  }
}


/* parses command line arguments and translates them to a configuration struct */
static struct conf *loadconf(int argc, char **argv) {
  int i;
  struct mdr_zip_item z;
  struct conf *cfg = NULL;

  cfg = malloc(sizeof(*cfg));
  if (cfg == NULL) {
    mdr_coutraw_puts("ERROR: OUT OF MEMORY");
    goto CLEANUP;
  }
  bzero(cfg, sizeof(*cfg));

  /* preset quizz limit to 30 questions */
  cfg->qizlimit = 30;

  /* read command-line arguments */
  for (i = 1; i < argc; i++) {
    /* handle switches */
    if (argv[i][0] == '/') {
      if (strcmp(argv[i] + 1, "?") == 0) {
        printhelp();
        goto CLEANUP;
      } else if (strcmp(argv[i] + 1, "nosound") == 0) {
        cfg->nosound = 1;
      } else if (strstr(argv[i] + 1, "max=") == argv[i] + 1) {
        cfg->qizlimit = atoi(argv[i] + 5);
      } else {
        mdr_coutraw_puts("ERROR: invalid switch");
      }
      continue;
    }
    /* else it's probably a QIZ file */
    if (cfg->qizfname != NULL) {
      mdr_coutraw_puts("ERROR: Too many arguments");
      return(NULL);
    }
    cfg->qizfname = argv[i];
  }

  /* do I have a qiz file at all? */
  if (cfg->qizfname == NULL) cfg->qizfname = "DEFAULT.QIZ";

  /* open qiz file */
  cfg->fd = fopen(cfg->qizfname, "rb");
  if (cfg->fd == NULL) {
    /* on error try again, appending a ".qiz" extension */
    if (strlen(cfg->qizfname) < 127 - 4) {
      char fname[128];
      sprintf(fname, "%s.qiz", cfg->qizfname);
      cfg->fd = fopen(fname, "rb");
    }
    /* still fail? then quit with error */
    if (cfg->fd == NULL) {
      mdr_coutraw_puts("ERROR: failed to open file:");
      mdr_coutraw_puts(cfg->qizfname);
      goto CLEANUP;
    }
  }

  /* load list of questions */
  bzero(&z, sizeof(z));
  for (;;) {
    int res;

    res = mdr_zip_iter(&z, cfg->fd);
    if (res == 1) break; /* end of archive */
    if (res != 0) {
      mdr_coutraw_puts("ERROR: invalid qiz file");
      goto CLEANUP;
    }

    if (z.flags & ZIP_FLAG_ISADIR) continue; /* skip directories */
    if (z.flags & ZIP_FLAG_ENCRYPTED) continue; /* skip encrypted entries */
    if (z.compmethod != ZIP_METH_STORE) continue; /* skip compressed entries */
    /* is it a title.pcx ? */
    if (fnmatch("title.pcx", z.fname, FNM_IGNORECASE) == 0) {
      cfg->titlepcx_offset = z.dataoffset;
      cfg->titlepcx_len = z.compressedfilelen;
      continue;
    }
    /* skip non-cfg files */
    if (fnmatch("*.CFG", z.fname, FNM_IGNORECASE) != 0) continue;

    /* alloc a new entry in list */
    cfg = realloc(cfg, sizeof(*cfg) + sizeof(struct qizlist) * (cfg->qizcount + 1));
    if (cfg == NULL) {
      mdr_coutraw_puts("ERROR: OUT OF MEMORY");
      goto CLEANUP;
    }
    bzero(&(cfg->list[cfg->qizcount]), sizeof(struct qizlist));
    cfg->list[cfg->qizcount].off = z.dataoffset;
    cfg->list[cfg->qizcount].len = z.compressedfilelen;

    cfg->qizcount++;
  }

  if (cfg->qizcount == 0) {
    mdr_coutraw_puts("ERROR: FOUND NO CFG FILES IN QIZ ARCHIVE");
    goto CLEANUP;
  }
  if ((cfg->qizlimit == 0) || (cfg->qizlimit > cfg->qizcount)) cfg->qizlimit = cfg->qizcount;

  return(cfg);

  CLEANUP:
  if (cfg != NULL) {
    if (cfg->fd != NULL) fclose(cfg->fd);
    free(cfg);
  }

  return(NULL);
}


static unsigned short losuj_cwiczenie(const struct conf *cfg) {
  unsigned short nonvalidated = 0xffff;
  unsigned short i;

  /* znajdz nastepne nieodpowiedziane pytanie */
  for (i = 0; i < cfg->qizlimit; i++) {
    /* skip validated questions */
    if (cfg->list[i].flags & QIZFLAG_VALIDATED) continue;
    /* never hinted? use it now */
    if ((cfg->list[i].flags & QIZFLAG_HINTED) == 0) return(i);
    /* zapamietaj niezatwierdzone pytanie */
    if (nonvalidated == 0xffff) nonvalidated = i;
  }

  /* jesli znalazlem jakies niezatwierdzone pytanie to dawaj */
  if (nonvalidated != 0xffff) return(nonvalidated);
  return(0);
}


/* shuffles the qiz list so qizlimit random questions are moved to the top of the list */
static void shuffleqiz(struct conf *cfg) {
  unsigned short i;
  unsigned short entry;
  struct qizlist buf;
  for (i = 0; i < cfg->qizlimit; i++) {
    /* losuj cwiczenie dla pozycji i */
    entry = i + (rand() % (cfg->qizcount - i)); /* entry = i ... qizcount-1 */
    /* nic nie rob jesli juz na dobrej pozycji */
    if (entry == i) continue;
    /* zamien pozycje entry oraz i miejscami */
    memcpy(&buf, &(cfg->list[i]), sizeof(struct qizlist));
    memcpy(&(cfg->list[i]), &(cfg->list[entry]), sizeof(struct qizlist));
    memcpy(&(cfg->list[entry]), &buf, sizeof(struct qizlist));
  }
}


/* trims trailing spaces and returns new len of s */
static unsigned short rtrimspc(char *s) {
  unsigned short i;
  unsigned short newlen = 0;
  for (i = 0; s[i] != 0; i++) {
    if (s[i] != ' ') newlen = i + 1;
  }
  s[newlen] = 0;
  return(newlen);
}


static void set_ui_pal(const struct pcx_hdr *monkey_hdr) {
  unsigned char i;
  for (i = 0; i < 16; i++) {
    video_setpalette(i + 240, monkey_hdr->pal[i].r >> 2, monkey_hdr->pal[i].g >> 2, monkey_hdr->pal[i].b >> 2);
  }
}


static void display_results(const struct conf *cfg, struct mbuf *mem, struct video_handler *vid) {
  char r[16];
  FILE *datfd;
  unsigned long wynik;
  enum MONKEY_FACE m;
  unsigned long foff;
  unsigned short fsiz;

  datfd = locateasset("OVER.PCX", &foff, &fsiz);
  if (datfd != NULL) {
    fadeout(mem->pal);
    pcx_display_fullscreen(datfd, foff, fsiz, 239, 0, mem->pal);
    fclose(datfd);
    set_ui_pal(&(mem->monkey_hdr)); /* revert UI colors that were overwritten by fadeout() */
  }
  wynik = cfg->stats_answers_good * 100 / cfg->stats_answers_total;
  sprintf(r, "%u%%", (unsigned short)wynik);
  vga_outstr(18, 9, r, PALID_BLACK, 0xff, 1);

  /* wskazowka */
  {
    unsigned long t;
    unsigned short rad, xx, yy;
    t = cfg->stats_answers_good;
    t *= 3142 - 80; /* -80 aby wskazowka byla lekko nachylona w pozycji 0 i 100% */
    t /= cfg->stats_answers_total;
    rad = t + 40;
    xx = 160 - trigint_cos(rad) / 12;
    yy = 124 - trigint_sin(rad) / 14;

    {
      unsigned char orx, ory;
      for (ory = 124-1; ory <= 124+1; ory++) {
        for (orx = 160-1; orx <= 160+1; orx++) {
          video_line(vid, orx, ory, xx, yy, PALID_BLACK_LIGHT);
        }
      }
      video_line(vid, 160+2, 124, xx, yy, PALID_BLACK_LIGHT);
      video_line(vid, 160-2, 124, xx, yy, PALID_BLACK_LIGHT);
      video_line(vid, 160, 124+2, xx, yy, PALID_BLACK_LIGHT);
      video_line(vid, 160, 124-2, xx, yy, PALID_BLACK_LIGHT);
    }
  }

  /* malpia minka zalezy od wyniku testu */

  if (wynik >= 95) {
    m = MONKEY_GRIN;
  } else if (wynik >= 85) {
    m = MONKEY_TONGUE;
  } else if (wynik >= 70) {
    m = MONKEY_BASIC;
  } else if (wynik >= 50) {
    m = MONKEY_UNSURE;
  } else if (wynik >= 40) {
    m = MONKEY_SAD;
  } else if (wynik >= 30) {
    m = MONKEY_SURPRISED;
  } else {
    m = MONKEY_OOUCH;
  }

  display_monkey_custom(MONKEY_BASIC, &(mem->monkey_hdr), mem->monkeybuf, 2, 250, 130);
  display_monkey_custom(m, &(mem->monkey_hdr), mem->monkeybuf, 2, 250, 130);
}


/* check whether a 64K page is crossed for a given seg:off pointer with len
 * bytes. Returns a pointer that is guaranteed not to cross such page.
 * This is useful when allocating buffers for DMA transfers. */
static void far *detect64kcrossing(void far *ptr, unsigned short len) {
  unsigned long physaddr, physaddr2;

  /* convert seg:off to physical address (eg. 2A83:FA9C becomes 0x0003A2CC) */
  physaddr = FP_SEG(ptr);
  physaddr <<= 4;
  physaddr += FP_OFF(ptr);

  /* compute a second physical address after len bytes */
  physaddr2 = physaddr + len;

  /* compare the highest 16 bits of both addresses. if equal, then no crossing
   * occurred. */
  physaddr2 ^= physaddr;
  physaddr2 >>= 16;
  if (physaddr2 == 0) return(ptr);

  /* otherwise increment physaddr and convert it to a segment address */
  physaddr >>= 16;
  physaddr++;
  physaddr <<= 12;
  return(MK_FP(physaddr, 0));
}


/* checks the answer against a list of possibilities, return 0 on match.
 * sollist is a list of strings terminated with an empty string. */
static int checkanswer(const char *sollist, const char *answer) {
  while (*sollist != 0) {
    if (strcasecmp(sollist, answer) == 0) return(0);
    sollist += strlen(sollist) + 1;
  }
  return(-1);
}


static void display_qiztext(FILE *fd, unsigned long off, unsigned short len) {
  unsigned char line = 0, col = 1;
  unsigned char b;
  fseek(fd, off, SEEK_SET);
  for (; len != 0; len--) {
    b = fgetc(fd);
    if (b == '\r') continue;
    if (b == '\n') {
      line++;
      col = 1;
      continue;
    }
    if ((line < 10) && (col <= 39)) vga_outchar(col++, line, b, PALID_BLACK, PALID_WHITE, 1);
  }
}


#define SOUNDBUFSZ 32000
#define SOUNDBUFSZ_BIG (SOUNDBUFSZ * 2)

int main(int argc, char **argv) {
  unsigned short frazalen;
  char fraza[64];
  int i;
  struct sbdigi_ctx *sbctx = NULL;
  struct video_handler *vid = NULL;
  int cwiczenie;
  unsigned char *soundbuf_big = NULL;
  unsigned char *soundbuf;
  unsigned char podpowiedz;
  struct conf *cfg = NULL;
  struct qiz q;
  struct mbuf *mem = NULL;

  cfg = loadconf(argc, argv);
  if (cfg == NULL) goto QUIT;

  /* memory allocation for general-purpose data buffer */
  mem = malloc(sizeof(struct mbuf));
  if (mem == NULL) {
    mdr_coutraw_puts("ERROR: out of memory");
    goto QUIT;
  }

  /* VGA check */
  if (!video_detectvga()) {
    mdr_coutraw_puts("ERROR: no VGA detected");
    goto QUIT;
  }

  /* init sound system (best effort) */
  if (cfg->nosound != 0) goto SOUND_INIT_DONE;

  soundbuf_big = malloc(SOUNDBUFSZ_BIG);
  if (soundbuf_big == NULL) goto SOUND_INIT_FAILED;

  /* SB requires a buffer that does not cross any 64K page since DMA
   * transfers are unable to cross such boundary */
  soundbuf = detect64kcrossing(soundbuf_big, SOUNDBUFSZ);
  if ((detect64kcrossing(soundbuf, SOUNDBUFSZ) != soundbuf)
   || (soundbuf + SOUNDBUFSZ > soundbuf_big + SOUNDBUFSZ_BIG)) {
    mdr_coutraw_puts("DMA BUFFER INIT FAILURE");
    goto QUIT;
  }

  sbctx = sbdigi_init(getenv("BLASTER"));
  if (sbctx == NULL) goto SOUND_INIT_FAILED;

  goto SOUND_INIT_DONE;

  SOUND_INIT_FAILED:
  if (sbctx != NULL) {
    sbdigi_quit(sbctx);
    sbctx = NULL;
  }
  if (soundbuf_big != NULL) free(soundbuf_big);
  cfg->nosound = 1;

  SOUND_INIT_DONE:

  /* load Monkey data */
  if (monkey_load(&(mem->monkey_hdr), mem->monkeybuf, sizeof(mem->monkeybuf)) != 0) {
    mdr_coutraw_puts("ERROR: failed to load data file");
    goto QUIT;
  }

  vid = video_open(0x13, 0);
  if (vid == NULL) {
    mdr_coutraw_puts("ERROR: video_open() failed");
    goto QUIT;
  }

  /* activate 8x16 fonts instead of the default ugly 8x8 glyphs, this means */
  /* that an 8x16 bitmap font will be loaded under the 0x43 vector */
#ifndef TESTBUILD
  _asm {
    mov ax, 0x1124
    mov bl, 1
    int 0x10
  }
#endif

  /* random generator initialization */
  randinit();

  /* display the intro screen, if found */
  if (cfg->titlepcx_len > 0) {
    if (pcx_display_fullscreen(cfg->fd, cfg->titlepcx_offset, cfg->titlepcx_len, 255, 1, mem->pal) == 0) {
      keyb_getkey();
      fadeout(mem->pal);
      video_cls(vid, 0);
    }
  }


  REPEAT_TEST:

  /* set up VGA palette with 16 UI colors provided by Monkey data.
   * all UI colors land in the last 16 positions of the VGA palette. */
  set_ui_pal(&(mem->monkey_hdr));
  memcpy(mem->pal + (240*3), mem->monkey_hdr.pal, 16*3);

  /* clear out counters and flags on questions */
  for (i = 0; i < cfg->qizcount; i++) cfg->list[i].flags = 0;
  cfg->stats_answers_total = 0;
  cfg->stats_answers_good = 0;

  /* shuffle the qizlist */
  shuffleqiz(cfg);

  NEXT_TRY:
  bzero(fraza, sizeof(fraza));
  frazalen = 0;

  /* losuj cwiczenie (nieodpowiedziane) */
  cwiczenie = losuj_cwiczenie(cfg);
  podpowiedz = 0;

  /* parse the CFG file and fill list node accordingly */
  if (load_qiz_item(&q, cfg, cwiczenie, soundbuf, SOUNDBUFSZ, mem->solution, sizeof(mem->solution)) != 0) {
    vga_outstr(1, 1, "CFG error in the QIZ file", PALID_WHITE, PALID_BLACK, 0);
    keyb_getkey();
    goto QUIT;
  }

  if (q.pcxlen != 0) {
    if (pcx_display_fullscreen(cfg->fd, q.pcxoffset, q.pcxlen, 239, 0, mem->pal) != 0) {
      vga_outstr(1, 1, "Failed to load PCX file", PALID_BROWN, PALID_WHITE, 0);
    }
  } else {
    video_rectfill(vid, 0, 0, 320, 168, PALID_BEIGE_LIGHT);
  }

  if (q.txtlen != 0) {
    display_qiztext(cfg->fd, q.txtoffset, q.txtlen);
  }

  display_monkey(MONKEY_BASIC, &(mem->monkey_hdr), mem->monkeybuf);

  video_rect(vid, 0, 168, 320, 32, PALID_BROWN);
  { /* corners of the input box are rounded by copying pixels above them */
    char unsigned far *vga = MK_FP(0xA000, 0);
    vga[168 * 320] = vga[167 * 320];
    vga[168 * 320 + 319] = vga[167 * 320 + 319];
  }

  /* fill background of the input box */
  video_rectfill(vid, 1, 169, 318, 30, PALID_WHITE);

  /* render (empty) success meter */
  video_rect(vid, 0, 198, 320, 2, PALID_GRAY_LIGHT);

  /* flush keyboard buffer */
  keyb_flush();

  /* play audio cue (if any and if sound enabled) */
  if ((sbctx != NULL) && (q.wavlen > 0)) {
    mdr_bios_tickswait(250 / 55);
    play_audiofile(sbctx, soundbuf, SOUNDBUFSZ, cfg, &q);
  }

  for (;;) {
    int k;

    /* display user's string (last char only, for performance, rest was already displayed earlier) */
    if (fraza[0] != 0) vga_outchar(strlen(fraza), 11, fraza[strlen(fraza) - 1], PALID_BLACK, PALID_WHITE, 0);
    /* display a "cursor" if there is any place left and display a trailing blank */
    if (frazalen < 38) {
      vga_outchar(frazalen + 1, 11, '_', PALID_GRAY_DARK, PALID_WHITE, 0);
      if (frazalen < 37) vga_outchar(frazalen + 2, 11, ' ', PALID_BLACK, PALID_WHITE, 0);
    }

    /* render success meter */
    video_rect(vid, 0, 198, 320 * cfg->stats_answers_total / cfg->qizlimit, 2, PALID_GREEN);

    /* read key (ignore extended control keys) */
    do {
      k = keyb_getkey();
    } while (k > 0xff);

    /* escape? */
    if (k == 0x1B) goto QUIT;

    /* enter? */
    if ((k == 0x0D) && (frazalen > 0)) {
      /* flush keyboard buffer */
      keyb_flush();
      /* strip trailing spaces */
      frazalen = rtrimspc(fraza);
      /* good answer */
      if (checkanswer(mem->solution, fraza) == 0) {
        if (podpowiedz == 0) {
          /* wyswietl jedna z kilku minek */
          switch (rand() & 1) {
            case 0:
              display_monkey(MONKEY_GRIN, &(mem->monkey_hdr), mem->monkeybuf);
              break;
            case 1:
              display_monkey(MONKEY_TONGUE, &(mem->monkey_hdr), mem->monkeybuf);
              break;
          }
          /* jeśli nie było nigdy podpowiedzi to zalicz */
          if ((cfg->list[cwiczenie].flags & QIZFLAG_HINTED) == 0) cfg->stats_answers_good++;
          /* mark the question as validated */
          cfg->list[cwiczenie].flags |= QIZFLAG_VALIDATED;
          cfg->stats_answers_total++;
        } else { /* dobrze, ale z podpowiedzia */
          display_monkey(MONKEY_BASIC, &(mem->monkey_hdr), mem->monkeybuf);
          /* flag question as 'hinted' - unless already hinted, then validate
           * it: no need to harass the user if his porch light ain't on */
          if (cfg->list[cwiczenie].flags & QIZFLAG_HINTED) {
            cfg->list[cwiczenie].flags |= QIZFLAG_VALIDATED;
            cfg->stats_answers_total++;
          }
          cfg->list[cwiczenie].flags |= QIZFLAG_HINTED;
        }
        if (sbctx != NULL) beep_goodanswer(sbctx, soundbuf);
        mdr_bios_tickswait(1000 / 55);
        if (cfg->stats_answers_total == cfg->qizlimit) break;
        goto NEXT_TRY;
      }
      /* bad answer */
      switch (rand() & 3) {
        case 0:
          display_monkey(MONKEY_SAD, &(mem->monkey_hdr), mem->monkeybuf);
          break;
        case 1:
          display_monkey(MONKEY_UNSURE, &(mem->monkey_hdr), mem->monkeybuf);
          break;
        case 2:
          display_monkey(MONKEY_OOUCH, &(mem->monkey_hdr), mem->monkeybuf);
          break;
        case 3:
          display_monkey(MONKEY_SURPRISED, &(mem->monkey_hdr), mem->monkeybuf);
          break;
      }
      if (sbctx != NULL) {
        beep_badanswer(sbctx, soundbuf);
        if (q.wavlen > 0) {
          mdr_bios_tickswait(900 / 55);
          play_audiofile(sbctx, soundbuf, SOUNDBUFSZ, cfg, &q);
        }
      }
      /* print out the solution and remember there's a hint going on */
      vga_outstr(20 - strlen(mem->solution) / 2, 9, mem->solution, PALID_BROWN, PALID_WHITE, 0);
      podpowiedz = 1;
    }

    /* backspace? */
    if ((k == 0x08) && (frazalen > 0)) {
      fraza[--frazalen] = 0;
    }

    if (((k > ' ') && (frazalen < 38)) || ((k == ' ') && (frazalen < 38) && (frazalen > 0))) {
      fraza[frazalen++] = k;
      fraza[frazalen] = 0;
    }
  }

  /* display results */
  display_results(cfg, mem, vid);
  if (keyb_getkey() == 'r') goto REPEAT_TEST;

  QUIT:

  if (vid != NULL) {
    if (mem != NULL) fadeout(mem->pal);
    video_cls(vid, 0);
    video_close(vid);
  }

  if (sbctx != NULL) sbdigi_quit(sbctx);
  if (mem != NULL) free(mem);
  if (soundbuf_big != NULL) free(soundbuf_big);

  if (cfg != NULL) {
    if (cfg->qizlimit > 0) {
      sprintf(fraza, "Last score: %u/%u (%u%%)", cfg->stats_answers_good, cfg->qizlimit, cfg->stats_answers_good * 100 / cfg->qizlimit);
      mdr_coutraw_puts(fraza);
      mdr_coutraw_puts("");
      mdr_coutraw_puts("See you later!");
    }
    if (cfg->fd != NULL) fclose(cfg->fd);
    free(cfg);
  }

  return(0);
}
